@import ObjectiveC.runtime;
@import MachO.loader;
@import Darwin.POSIX.dlfcn;

Method getMethod(NSString* className,NSString* selName,BOOL isInstance)
{
	Class class=NSClassFromString(className);
	SEL sel=NSSelectorFromString(selName);
	
	if(isInstance)
	{
		return class_getInstanceMethod(class,sel);
	}
	return class_getClassMethod(class,sel);
}

BOOL swizzle(NSString* realClass,NSString* fakeClass,NSString* realSel,NSString* fakeSel,BOOL isInstance)
{
	Method realMethod=getMethod(realClass,realSel,isInstance);
	Method fakeMethod=getMethod(fakeClass,fakeSel,isInstance);
	
	if(!realMethod||!fakeMethod)
	{
		return false;
	}
	
	method_exchangeImplementations(realMethod,fakeMethod);
	
	return true;
}

BOOL swizzleSimple(NSString* className,NSString* selName,BOOL isInstance)
{
	NSString* fakeSelName=[@"fake_" stringByAppendingString:selName];
	return swizzle(className,@"Fake",selName,fakeSelName,isInstance);
}

void swizzlePrinting(NSString* realClass,NSString* fakeClass,NSString* realSel,NSString* fakeSel,BOOL isInstance)
{
	trace(@"swizzle %@.%@ %@.%@: %@",realClass,realSel,fakeClass,fakeSel,swizzle(realClass,fakeClass,realSel,fakeSel,isInstance)?@"success":@"failure");
}

void swizzleSimplePrinting(NSString* className,NSString* selName,BOOL isInstance)
{
	trace(@"swizzleSimple %@.%@: %@",className,selName,swizzleSimple(className,selName,isInstance)?@"success":@"failure");
}

// https://opensource.apple.com/source/dyld/dyld-832.7.3/include/mach-o/dyld-interposing.h

#define interposeStatic(_replacement,_replacee) __attribute__((used)) static struct{const void* replacement;const void* replacee;} _interpose_##_replacee __attribute__ ((section ("__DATA,__interpose"))) = {(const void*)(unsigned long)&_replacement,(const void*)(unsigned long)&_replacee};

// http://blog.darlinghq.org/2018/07/mach-o-linking-and-loading-tricks.html
// https://chromium.googlesource.com/chromium/src/+/refs/tags/90.0.4411.1/media/audio/mac/coreaudio_dispatch_override.cc

// TODO: doesn't seem to work...

typedef struct
{
	void* replacement;
	void* replacee;
} interposeTuple;

extern void dyld_dynamic_interpose(const struct mach_header*,const interposeTuple replacements[],size_t count);

void interposeDynamic(void* replacement,void* replacee)
{
	Dl_info info;
	dladdr(replacee,&info);
	
	interposeTuple tuple={replacement,replacee};
	
	dyld_dynamic_interpose(info.dli_fbase,&tuple,1);
}